/*
 * gimmix-io.c -- University of Applied Sciences Giessen
 *                MMIX-IO implementation
 */


//<Preprocessor macros 2>
#include <stdio.h>
#include <stdlib.h>

#include "gimmix-arith.h"
#include "gimmix-io.h"
#include "gimmix-sim.h"


#ifndef FILENAME_MAX
#define FILENAME_MAX		256
#endif


//#define sign_bit		((unsigned) 0x80000000)


//<Type definitions 3>
//typedef unsigned int tetra;
//typedef struct {
//  tetra h;
//  tetra l;
//} octa;

typedef struct {
  FILE *fp;	/* file pointer */
  int mode;	/* 1[read OK] + 2[write OK] + 4[binary] + 8[readwrite] */
} sim_file_info;


//<External subroutines 4>
//extern char stdin_chr(void);
//extern int mmgetchars(char *buf, int size, octa addr, int stop);
//extern void mmputchars(unsigned char *buf, int size, octa addr);
//extern octa oplus(octa, octa);
//extern octa ominus(octa, octa);
//extern octa incr(octa, int);
//extern octa zero_octa;
//extern octa neg_one;


//<Global variables 6>
sim_file_info sfile[256];

char *mode_string[] = {
  "r", "w", "rb", "wb", "w+b"
};

int mode_code[] = {
  0x1, 0x2, 0x5, 0x6, 0xf
};

char *trip_warning[] = {
  "TRIP",
  "integer divide check",
  "integer overflow",
  "float-to-fix overflow",
  "invalid floating point operation",
  "floating point overflow",
  "floating point underflow",
  "floating point division by zero",
  "floating point inexact"
};


//<Subroutines 7>
void mmix_io_init(void) {
  sfile[0].fp = stdin;
  sfile[0].mode = 1;
  sfile[1].fp = stdout;
  sfile[1].mode = 2;
  sfile[2].fp = stderr;
  sfile[2].mode = 2;
}


octa mmix_fopen(unsigned char handle, octa name, octa mode) {
  char name_buf[FILENAME_MAX];

  if (mode.h != 0 || mode.l > 4) {
    sfile[handle].mode = 0;
    return neg_one;
  }
  if (mmgetchars(name_buf, FILENAME_MAX, name, 0) == FILENAME_MAX) {
    sfile[handle].mode = 0;
    return neg_one;
  }
  if (sfile[handle].mode != 0 && handle > 2) {
    fclose(sfile[handle].fp);
  }
  sfile[handle].fp = fopen(name_buf, mode_string[mode.l]);
  if (sfile[handle].fp == NULL) {
    sfile[handle].mode = 0;
    return neg_one;
  }
  sfile[handle].mode = mode_code[mode.l];
  /* success */
  return zero_octa;
}


void mmix_fake_stdin(FILE *f) {
  /* f should be open in mode "r" */
  sfile[0].fp = f;
}


octa mmix_fclose(unsigned char handle) {
  if (sfile[handle].mode == 0) {
    return neg_one;
  }
  if (handle > 2 && fclose(sfile[handle].fp) != 0) {
    return neg_one;
  }
  sfile[handle].mode = 0;
  /* success */
  return zero_octa;
}


octa mmix_fread(unsigned char handle, octa buffer, octa size) {
  unsigned char *buf;
  int n;
  octa o;
  unsigned char *p;

  o = neg_one;
  if ((sfile[handle].mode & 0x1) == 0) {
    return ominus(o, size);
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x2;
  }
  if (size.h != 0) {
    return ominus(o, size);
  }
  buf = (unsigned char *) calloc(size.l, sizeof(char));
  if (buf == NULL) {
    return ominus(o, size);
  }
  //<Read n <= size.l characters into buf 13>
  if (sfile[handle].fp == stdin) {
    for (p = buf, n = size.l; p < buf + n; p++) {
      *p = stdin_chr();
    }
  } else {
    clearerr(sfile[handle].fp);
    n = fread(buf, 1, size.l, sfile[handle].fp);
    if (ferror(sfile[handle].fp)) {
      return ominus(o, size);
    }
  }
  mmputchars(buf, n, buffer);
  free(buf);
  o.h = 0;
  o.l = n;
  return ominus(o, size);
}


octa mmix_fgets(unsigned char handle, octa buffer, octa size) {
  char buf[256];
  int n, s;
  char *p;
  octa o;
  int eof;

  eof = 0;
  if ((sfile[handle].mode & 0x1) == 0) {
    return neg_one;
  }
  if (size.l == 0 && size.h == 0) {
    return neg_one;
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x2;
  }
  size = incr(size, -1);
  o = zero_octa;
  while (1) {
    //<Read n < 256 characters into buf 15>
    s = 255;
    if (size.l < s && size.h == 0) {
      s = size.l;
    }
    if (sfile[handle].fp == stdin) {
      for (p = buf, n = 0; n < s; ) {
        *p = stdin_chr();
        n++;
        if (*p++ == '\n') {
          break;
        }
      }
    } else {
      if (fgets(buf, s + 1, sfile[handle].fp) == NULL) {
        return neg_one;
      }
      eof = feof(sfile[handle].fp);
      for (p = buf, n = 0; n < s; ) {
        if (*p == '\0' && eof) {
          break;
        }
        n++;
        if (*p++ == '\n') {
          break;
        }
      }
    }
    *p = '\0';
    mmputchars(buf, n + 1, buffer);
    o = incr(o, n);
    size = incr(size, -n);
    if ((n != 0 && buf[n - 1] == '\n') ||
        (size.l == 0 && size.h == 0) ||
        eof) {
      return o;
    }
    buffer = incr(buffer, n);
  }
}


octa mmix_fgetws(unsigned char handle, octa buffer, octa size) {
  char buf[256];
  int n, s;
  char *p;
  octa o;
  int eof;

  eof = 0;
  if ((sfile[handle].mode & 0x1) == 0) {
    return neg_one;
  }
  if (size.l == 0 && size.h == 0) {
    return neg_one;
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x2;
  }
  buffer.l &= -2;
  size = incr(size, -1);
  o = zero_octa;
  while (1) {
    //<Read n < 128 wyde characters into buf 17>
    s = 127;
    if (size.l < s && size.h == 0) {
      s = size.l;
    }
    if (sfile[handle].fp == stdin) {
      for (p = buf, n = 0; n < s; ) {
        *p++ = stdin_chr();
        *p++ = stdin_chr();
        n++;
        if (*(p - 1) == '\n' && *(p - 2) == '\0') {
          break;
        }
      }
    } else {
      for (p = buf, n = 0; n < s; ) {
        if (fread(p, 1, 2, sfile[handle].fp) != 2) {
          eof = feof(sfile[handle].fp);
          if (eof == 0) {
            return neg_one;
          }
          break;
        }
        n++;
        p += 2;
        if (*(p - 1) == '\n' && *(p - 2) == '\0') {
          break;
        }
      }
    }
    *(p + 0) = '\0';
    *(p + 1) = '\0';
    mmputchars(buf, 2 * n + 2, buffer);
    o = incr(o, n);
    size = incr(size, -n);
    if ((n != 0 && buf[2 * n - 1] == '\n' && buf[2 * n - 2] == '\0') ||
        (size.l == 0 && size.h == 0) ||
        eof) {
      return o;
    }
    buffer = incr(buffer, 2 * n);
  }
}


octa mmix_fwrite(unsigned char handle, octa buffer, octa size) {
  char buf[256];
  int n;

  if ((sfile[handle].mode & 0x2) == 0) {
    return ominus(zero_octa, size);
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x1;
  }
  while (1) {
    if (size.h != 0 || size.l >= 256) {
      n = mmgetchars(buf, 256, buffer, -1);
    } else {
      n = mmgetchars(buf, size.l, buffer, -1);
    }
    size = incr(size, -n);
    if (fwrite(buf, 1, n, sfile[handle].fp) != n) {
      return ominus(zero_octa, size);
    }
    fflush(sfile[handle].fp);
    if (size.l == 0 && size.h == 0) {
      return zero_octa;
    }
    buffer = incr(buffer, n);
  }
}


octa mmix_fputs(unsigned char handle, octa string) {
  char buf[256];
  int n;
  octa o;

  o = zero_octa;
  if ((sfile[handle].mode & 0x2) == 0) {
    return neg_one;
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x1;
  }
  while (1) {
    n = mmgetchars(buf, 256, string, 0);
    if (fwrite(buf, 1, n, sfile[handle].fp) != n) {
      return neg_one;
    }
    o = incr(o, n);
    if (n < 256) {
      fflush(sfile[handle].fp);
      return o;
    }
    string = incr(string, n);
  }
}


octa mmix_fputws(unsigned char handle, octa string) {
  char buf[256];
  int n;
  octa o;

  o = zero_octa;
  if ((sfile[handle].mode & 0x2) == 0) {
    return neg_one;
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode &= ~0x1;
  }
  while (1) {
    n = mmgetchars(buf, 256, string, 1);
    if (fwrite(buf, 1, n, sfile[handle].fp) != n) {
      return neg_one;
    }
    o = incr(o, n >> 1);
    if (n < 256) {
      fflush(sfile[handle].fp);
      return o;
    }
    string = incr(string, n);
  }
}


octa mmix_fseek(unsigned char handle, octa offset) {
  if ((sfile[handle].mode & 0x4) == 0) {
    return neg_one;
  }
  if (sfile[handle].mode & 0x8) {
    sfile[handle].mode = 0xf;
  }
  if (offset.h & sign_bit) {
    if (offset.h != 0xffffffff || (offset.l & sign_bit) == 0) {
      return neg_one;
    }
    if (fseek(sfile[handle].fp, (int) offset.l + 1, SEEK_END) != 0) {
      return neg_one;
    }
  } else {
    if (offset.h != 0 || (offset.l & sign_bit) != 0) {
      return neg_one;
    }
    if (fseek(sfile[handle].fp, (int) offset.l, SEEK_SET) != 0) {
      return neg_one;
    }
  }
  return zero_octa;
}


octa mmix_ftell(unsigned char handle) {
  long x;
  octa o;

  if ((sfile[handle].mode & 0x4) == 0) {
    return neg_one;
  }
  x = ftell(sfile[handle].fp);
  if (x < 0) {
    return neg_one;
  }
  o.h = 0;
  o.l = x;
  return o;
}


void print_trip_warning(int n, octa loc) {
  if (sfile[2].mode & 0x2) {
    fprintf(sfile[2].fp,
            "Warning: %s at location %08x%08x\n",
            trip_warning[n], loc.h, loc.l);
  }
}
